# vim:set ts=2 sw=2 sts=2 et:
from rzn.obj.layers import Layers
from rzn.obj import units
from rzn.log import Logger

log = Logger(__name__)


class ToolOption(object):
  def __init__(self, tool, helptext):
    self.tool = tool
    self._choice = None
    self.name = None
    self.helptext = helptext

  def choose(self, i):
    self._choice = i
    self.tool.option_changed(self.name, self._choice)

  def choice(self):
    return self._choice

  def choices(self):
    return []

  def value(self):
    return self._choice

  def set_value(self, v):
    self._choice = v


class Choice(ToolOption):
  def __init__(self, tool, choices, helptext=""):
    ToolOption.__init__(self, tool, helptext)
    self._choices = choices
    self._choice = choices[0]

  def choices(self):
    return self._choices

  def cycle(self):
    idx = self._choices.index(self._choice)
    idx = (idx+1) % len(self._choices)
    self.choose(self._choices[idx])

  def set_value(self, v):
    if v not in self._choices:
      log.error("Invalid choice for tool option '%s' (should be in %s)" % (
        v, self._choices))
    ToolOption.set_value(self, v)


class Scale(ToolOption):
  def __init__(self, tool, helptext=""):
    ToolOption.__init__(self, tool, helptext)
    self._choices = [0.75+0.05*i for i in range(11)]
    self._choice = 1.0

  def choice(self):
    return str(self._choice)

  def choices(self):
    return map(str,self._choices)

  def set_value(self, v):
    ToolOption.set_value(self, float(v))

  def value(self):
    return self._choice


class Layer(ToolOption):
  def __init__(self, tool, choice=None, layershow="all", helptext=""):
    ToolOption.__init__(self, tool, helptext)
    self.layercount = 0
    self.layershow = layershow
    self._checklayers()
    viewname = self.tool.view().name
    if choice:
      self._choice = choice
    else:
      self._choice = "Net" if viewname == "schematic" else "Border"

  def _checklayers(self):
    view = self.tool.view()
    if view.project().layercount() == self.layercount:
      return
    self.layercount = view.project().layercount()
    nums = Layers.get(view.name, self.layershow, view.project().layercount())
    self.layernames = sorted([Layers.names[n] for n in nums])

  def choices(self):
    self._checklayers()
    return self.layernames

  def choose(self, v):
    self.tool.view().render.setLayerVisible(getattr(Layers, v), True)
    ToolOption.choose(self, v)

  def value(self):
    return getattr(Layers, self._choice)

  def set_value(self, v):
    if v not in Layers.names:
      log.error("Invalid layer '%s' for tool option" % v)
      return
    self._choice = Layers.names[v]


class Dimension(ToolOption):
  def __init__(self, tool, defaultidx=0, helptext=""):
    ToolOption.__init__(self, tool, helptext)
    scalings = [0,2.5,5,10,25,100,250]
    self.invals = [int(units.inch(0.001) * v) for v in scalings]
    self.mmvals = [int(units.mm(0.01) * v) for v in scalings]
    self._defaultidx = defaultidx
    self._um = 0

  def _values(self):
    inch = self.tool.view().render.grid.is_inches()
    return self.invals if inch else self.mmvals

  def _umstr(self, um):
    inch = self.tool.view().render.grid.is_inches()
    scale = float(units.inch(1) if inch else units.mm(1))
    s = "%0.4f" % (um/scale)
    s = s.rstrip("0")
    if s.endswith("."):
      s += "0"
    return s

  def _checkval(self, um):
    inch = self.tool.view().render.grid.is_inches()
    if inch:
      if um not in self.invals:
        self.invals = sorted(self.invals+[um])
    else:
      if um not in self.mmvals:
        self.mmvals = sorted(self.mmvals+[um])

  def choice(self):
    if self._choice is None:
      val = self._values()[self._defaultidx]
      self._choice = self._umstr(val)
      self._um = val
    return self._umstr(self._um)

  def choices(self):
    self._checkval(self._um)
    return [self._umstr(v) for v in self._values()]+["..."]
  
  def choose(self, i):
    inch = self.tool.view().render.grid.is_inches()
    self._um = float(i)*float(units.inch(1) if inch else units.mm(1))
    self._checkval(self._um)
    ToolOption.choose(self, i)

  def value(self):
    if self._choice is None:
      val = self._values()[self._defaultidx]
      self._choice = self._umstr(val)
      self._um = val
    return self._um

  def set_value(self, v):
    if not isinstance(v, (int, float)):
      log.error("Invalid value '%s' for dimension tool option" % v)
    self._um = v
    self._choice = self._umstr(v)
