# vim:set ts=2 sw=2 sts=2 et:
from rzn.inputdev import mouse
from rzn.obj.objs import Arc
from .tool import Tool
from . import options
from .arcsupport import centre_from_3points
import math

(STYLE_SEM, # start/end/mid
 STYLE_CSE, # centre/start/end
 ) = range(2)

STYLES = ["Start/End/Mid", "Centre/Start/End"]


class ToolArc(Tool):
  name = "Arc"

  def __init__(self, tools):
    Tool.__init__(self, tools)
    self.ghost = None
    self.helptext = self.tr("Draw arcs, circles, and curved traces")
    self.add_option("Layer", options.Layer(self, helptext=self.tr("Set arc layer")))
    self.add_option("Width", options.Dimension(self, helptext=self.tr("Set arc line width")))
    self.add_option("Style", 
        options.Choice(self, STYLES,
                       helptext=self.tr("Set arc definition style")))

  def enter(self):
    self.ghost = None
    self.state = 0
    self._update_help()

  def _style(self):
    return STYLES.index(self.get_option("Style"))

  def option_changed(self, optname, value):
    Tool.option_changed(self, optname, value)
    if optname == "Style":
      if self.ghost:
        self.ghost.delete()
      self.ghost = None
      self.state = 0
    elif optname == "Width":
      if self.ghost:
        g = self.ghost
        self.ghost = self._mkarc(g.pos, g.ofsa, g.ofsb)
        g.delete()

  def exit(self):
    self.set_help("")
    if self.ghost:
      self.ghost.delete()

  def _mkarc(self, pos, ofsa, ofsb, ghost=False):
    view = self.view()
    # scale ofsb to match ofsa if necessary
    ra = math.hypot(*ofsa)
    if ra > 0:
      scaleb = math.hypot(*ofsb)/ra
      if scaleb != 1.0:
        ofsb = (int(ofsb[0]/scaleb), int(ofsb[1]/scaleb))
    a = view.design.make_obj(Arc, view.design, pos, ofsa, ofsb, ghost=ghost)
    a.layer = self.get_option("Layer")
    a.width = self.get_option("Width")
    return a

  def _update_help(self):
    style = self._style()
    helps = {"start" : self.tr("Click: set arc start point"),
             "end" : self.tr("Click: set arc end point"),
             "mid" : self.tr("Click: set arc center point"),
             "rad" : self.tr("Click: set arc radius")}
    if style == STYLE_CSE:
      if self.state == 0:   self.set_help(helps["mid"])
      elif self.state == 1: self.set_help(helps["start"])
      elif self.state == 2: self.set_help(helps["end"])
    elif style == STYLE_SEM:
      if self.state == 0:   self.set_help(helps["start"])
      elif self.state == 1: self.set_help(helps["end"])
      elif self.state == 2: self.set_help(helps["rad"])

  def _update_arc(self, x, y):
    style = self._style()
    if style == STYLE_CSE:
      if self.state == 1:
        ofs = (x-self.pt1[0], y-self.pt1[1])
        self.ghost.ofsb = ofs
        self.ghost.ofsa = ofs
      elif self.state == 2:
        self.ghost.ofsa = (self.pt2[0]-self.pt1[0], self.pt2[1]-self.pt1[1])
        self.ghost.ofsb = (x-self.pt1[0], y-self.pt1[1])
    elif style == STYLE_SEM:
      if self.state == 1:
        ax, ay = self.pt1
        bx, by = x, y
        pos = (ax+bx)/2, (ay+by)/2
        ofsa = ax-pos[0], ay-pos[1]
        ofsb = bx-pos[0], by-pos[1]
        self.ghost.pos = pos
        self.ghost.ofsa = ofsa
        self.ghost.ofsb = ofsb
      elif self.state == 2:
        centre = centre_from_3points(self.pt1, self.pt2, (x,y))
        if centre:
          ax, ay = self.pt1
          bx, by = self.pt2
          ofsa = ax-centre[0], ay-centre[1]
          ofsb = bx-centre[0], by-centre[1]
          if (bx-ax)*(y-ay) - (by-ay)*(x-ax) > 0:
            ofsa, ofsb = ofsb, ofsa
          self.ghost.pos = centre
          self.ghost.ofsa = ofsa
          self.ghost.ofsb = ofsb

  def _handle_press(self, pt):
    if self.state == 0:
      self.pt1 = pt
      self.ghost = self._mkarc(pt, (0,0), (0,0), ghost=True)
      self.state = 1
    elif self.state == 1:
      self.pt2 = pt
      self.state = 2
    elif self.state == 2:
      self._mkarc(self.ghost.pos, self.ghost.ofsa, self.ghost.ofsb)
      self.state = 0
      self.ghost.delete()
      self.ghost = None
    self._update_arc(*pt)

  def move(self, x, y, dx, dy):
    if self.ghost:
      self._update_arc(*self.view().abs_pos_grid())

  def press(self, x, y, button, mod):
    if button == mouse.LEFT:
      pt = self.view().abs_pos_grid()
      self._handle_press(pt)
      self._update_help()

