# vim:set ts=2 sw=2 sts=2 et:
from rzn.inputdev import mouse
from rzn.obj.objs import Segment
from .tool import Tool
from . import options

class ToolLine(Tool):
  name = "Line"
  objcls = Segment

  def __init__(self, tools):
    Tool.__init__(self, tools)
    self.helptext = self.tr("Draw lines and rectangles")
    sch = (self.view().name == "schematic")
    self.add_option("Layer", 
        options.Layer(self, "Symbol" if sch else "Border",
                      helptext=self.tr("Set line layer")))
    self.add_option("Width", 
        options.Dimension(self, helptext=self.tr("Set line width")))
    self.add_option("Style", 
        options.Choice(self, ["Right A", "Right B", "Direct"], 
                       helptext=self.tr("Set segment style")))
    self.ghost = []
  
  def option_changed(self, optname, value):
    Tool.option_changed(self, optname, value)
    if optname == "Layer":
      nghost = []
      while len(self.ghost):
        g = self.ghost.pop()
        g.delete()
        nghost.append(self._mkseg(g.pta, g.ptb, ghost=True, name="line"))
      self.ghost = nghost

  def enter(self):
    self.pts = []
    self.ghost = []
    self.set_help(self.tr("Click: start line"))
    self.view().render.setLayerVisible(self.get_option("Layer"), True)
    self.segs = []

  def _update_help(self):
    self.set_help(self.tr("Click: create line segment, Right-click: change draw style"))

  def _delete_ghosts(self):
    while len(self.ghost):
      self.ghost.pop().delete()
  
  def exit(self):
    self.set_help("")
    self._delete_ghosts()
  
  def move(self, x, y, dx, dy):
    if len(self.pts):
      pt = self.view().abs_pos_grid()
      self._delete_ghosts()
      self.ghost = self._mksegs(self.pts[-1], pt, ghost=True, name="line")

  def _mkseg(self, pta, ptb, ghost=False, name=None):
    view = self.view()
    s = view.design.make_obj(self.objcls, view.design, pta, ptb, ghost=ghost)
    s.layer = self.get_option("Layer")
    s.width = self.get_option("Width")
    if name:
      s.name = name
    return s
  
  def _mksegs(self, pta, ptb, ghost=False, name=None):
    if name is None:
      name = self.view().design.project.newNet()
    linetype = self.get_option("Style")
    if linetype == "Direct":
      ptm = pta
    elif linetype == "Right A":
      ptm = (ptb[0], pta[1])
    else:
      ptm = (pta[0], ptb[1])
    if (pta == ptm) or (ptb == ptm):
      return [self._mkseg(pta, ptb, ghost, name)]
    else:
      return [self._mkseg(pta, ptm, ghost, name), 
              self._mkseg(ptm, ptb, ghost, name)]

  def press(self, x, y, button, mod):
    view = self.view()

    if button == mouse.LEFT:
      pt = view.abs_pos_grid()
      if len(self.pts) == 0 or self.pts[-1] != pt:
        self.pts.append(pt)
        self._delete_ghosts()
      # if we have 2, make a segment
      if len(self.pts) == 2:
        pa, pb = self.pts
        self.pts.pop(0)
        s = self._mksegs(pa, pb, name="line")
        self.segs.extend(s)
      self._update_help()
    elif button == mouse.RIGHT:
      self.options["Style"].cycle()

