# vim:set ts=2 sw=2 sts=2 et:
from .tool import Tool
from rzn.inputdev import mouse
from rzn.obj.objs import Symbol, Footprint
from rzn.obj.design import Schematic
from rzn import support
from rzn.vc import vclib
from rzn.log import Logger
import os

log = Logger(__name__)

def update_libdeps(proj, lname, pname):
  path = support.partpath(lname, pname)
  librev = proj.getLibDep(lname)
  revs = [vclib.part_rev(path)]
  if librev:
    revs.append(librev)
  rev = vclib.latest_of_changesets(support.libpath(lname), revs)
  proj.setLibDep(lname, rev)
  return path

def makepart(proj, lname, pname, fpos, spos):
  path = update_libdeps(proj, lname, pname)
  part = proj.getPart(path, proj.layout)
  name = proj.newPartName(part.prefix)
  f = s = None
  if len(part.layout):
    f = proj.layout.make_obj(Footprint, proj.layout, fpos, lname, pname, name)
  if len(part.schematic):
    s = proj.schematic.make_obj(Symbol, proj.schematic, spos, lname, pname, name)
  return f, s, name

class ToolPart(Tool):
  name = "Part"

  def __init__(self, tools):
    Tool.__init__(self, tools)
    self.helptext = self.tr("Add a symbol and footprint to the design")
    self.ghost = None

  def enter(self):
    self.lname, self.pname = self.tools.show_part_selection()
    self.ghost = None
    if self.pname == "":
      self.tools.switchtool("Select")
    else:
      self._createpart()
  
  def exit(self):
    self.set_help("")
    if self.ghost:
      self.ghost.delete()
      self.ghost = None

  def move(self, x, y, dx, dy):
    if self.ghost:
      view = self.view()
      pt = view.abs_pos_grid()
      self.ghost.pos = pt

  def press(self, x, y, button, mod):
    if button == mouse.LEFT:
      self._place()
    elif button == mouse.RIGHT:
      if self.ghost:
        self.ghost.angle = (self.ghost.angle+90)%360

  def _place(self):
    proj = self.view().design.project
    if self.inschematic:
      fpos, spos = proj.layout.findFreeSpace(), self.ghost.pos
    else:
      fpos, spos = self.ghost.pos, proj.schematic.findFreeSpace()
    f, s, name = makepart(proj, self.lname, self.pname, fpos, spos)
    if self.inschematic:
      if f:
        f.pos = proj.layout.findFreeSpace(f.bb)
      if s:
        s.angle = self.ghost.angle
    else:
      if s:
        s.pos = proj.schematic.findFreeSpace(s.bb)
      if f:
        f.angle = self.ghost.angle
    proj.addDesignPart(s, f, name)

  def _createpart(self):
    view = self.view()
    proj = view.design.project
    if not (self.pname and 
            os.path.exists(support.partpath(self.lname, self.pname))):
      log.error("'%s' not found" % self.pname)
      self.tools.switchtool("Select")
      return
    self.set_help(self.tr("Click: place '%s', Right-click: rotate") % self.pname)
    self.inschematic = isinstance(view.design, Schematic)
    schematic, layout = proj.schematic, proj.layout
    pt = view.abs_pos_grid()
    update_libdeps(proj, self.lname, self.pname)
    if self.inschematic:
      self.ghost = schematic.make_obj(Symbol, schematic, pt, 
                             self.lname, self.pname, "??", ghost=True)
    else:
      self.ghost = layout.make_obj(Footprint, layout, pt, 
                          self.lname, self.pname, "??", ghost=True)
    self.view().selectObjects([self.ghost])
    self.move(0,0,0,0)

  def valid_for(self, view, part):
    return not part
