# vim:set ts=2 sw=2 sts=2 et:
from .tool import Tool
from rzn.obj.objs import (Segment, Pin, Pad, Symbol, Footprint, Text, Arc, Drill,
                          Polygon)
from rzn.obj.design import Schematic
from rzn.obj.loader import element_from_json
from rzn import support
import rzn


def handle_subpart(proj, s, namemap):
  if s["name"] not in namemap:
    path = support.partpath(s["lib"], s["part"])
    part = proj.getPart(path, proj.layout)
    namemap[s["name"]] = proj.newPartName(part.prefix)
  s["name"] = namemap[s["name"]]
  return s["name"]

# return where to paste object, either:
# - the current design view
# - the same view type as the copy (for footprints/symbols)
# - nowehere (like trying to paste a pin into a layout)
def choosebase(base, jsonobj, design):
  slmap = {
    "Pin":"s", "Net":"s",
    "Pad":"l", "Via":"l", "Trace":"l", "Drill":"l", "Polygon":"l",
  }
  curview = "s" if isinstance(design, Schematic) else "l"
  # don't paste if we're parented or the object isn't naturally pastable in 
  # the current view
  if ((curview not in slmap.get(jsonobj["_t"], "sl")) or ("_p" in jsonobj)):
    return None
  if (jsonobj["_t"] in ["Symbol", "Footprint", "Pin", "Pad", "Net", 
                       "Trace", "Via", "Drill"]):
    sch = design.project.schematic
    lay = design.project.layout
    return sch if isinstance(base, Schematic) else lay
  else:
    return design

class ToolPaste(Tool):
  name = "Paste"

  def enter(self):
    design = self.view().design
    items, opos = rzn.Razen.window().clipboard.getAll()
    if not opos:
      return self.tools.switchtool("Select")
    # offset from copy position
    pos = self.view().abs_pos_grid()
    ofsx, ofsy = pos[0]-opos[0], pos[1]-opos[1]
    if len(items) == 0:
      self.tools.switchtool("Select")
      return
    selitems = []
    # pre-move
    partnames = set()
    namemap = {}
    sch = design.project.schematic
    lay = design.project.layout
    revs = {}
    for base, s in items:
      mybase = choosebase(base, s, design)
      if mybase:
        # if this is a subpart rename it appropriately
        if s["_t"] == "Symbol" or s["_t"] == "Footprint":
          partnames.add(handle_subpart(design.project, s, namemap))
          revs[s["lib"]] = base.project.getLibDep(s["lib"])
        # create object from the json
        o = element_from_json(s, mybase, mybase, True)
        if mybase == design:
          selitems.append(o)
          if isinstance(o, Segment):
            o.pta = (o.pta[0]+ofsx, o.pta[1]+ofsy)
            o.ptb = (o.ptb[0]+ofsx, o.ptb[1]+ofsy)
          elif isinstance(o, (Pin, Pad, Symbol, Footprint, Text, 
                              Arc, Drill, Polygon)):
            o.pos = (o.pos[0]+ofsx, o.pos[1]+ofsy)
          else:
            raise Exception("unknown paste")
        else:
          o.pos = mybase.findFreeSpace()
    # update lib deps
    for lname, rev in revs.items():
      design.project.setLibDep(lname, rev)
    # update designparts
    for partname in partnames:
      design.project.addDesignPart(sch.findSubPartNamed(partname),
                                   lay.findSubPartNamed(partname),
                                   partname)
    if len(selitems):
      # ensure pasted layers are visible
      for o in selitems:
        if o.hasprop("layer"):
          self.view().render.setLayerVisible(o.layer, True)
      # select items for move and switch to move tool
      self.view().selectObjects(selitems)
      self.tools.switchtool("Move", False)
    else:
      self.tools.switchtool("Select")
  
  def valid_for(self, view, part):
    return False
