# vim:set ts=2 sw=2 sts=2 et:
from .tool import Tool
from rzn.inputdev import mouse
from rzn.obj.objs.common import (distPointToPoint, distPointToLine, 
                                 nearestPointOnLine)
from rzn.obj.objs.boundbox import BoundBox
from rzn.obj.objs import Via, Trace, Pad, Footprint, Arc, SubPart
from rzn.obj.layers import Layers
from rzn.obj import units
from rzn.render import helpers
from rzn.drc import drc
from rzn import config
from . import options

class ToolRoute(Tool):
  name = "Route"

  def __init__(self, tools):
    Tool.__init__(self, tools)
    self.helptext = self.tr("Draw traces for a selected net")
    self.lines = []
    self.pts = []
    self.endpoints = []
    self.ghosts = []
    self.drc = drc.DRC(self.view().design)
    self.netname = None
    self.add_option("Layer", options.Layer(self, layershow="copper",
                                           helptext=self.tr("Set routing layer")))
    self.add_option("Width", options.Dimension(self, defaultidx=2, 
                                               helptext=self.tr("Set trace width")))
    self.add_option("Style", 
        options.Choice(self, ["Trapezoid A", "Trapezoid B", "Arc A", "Arc B", "Direct"], 
                       helptext=self.tr("Set routing style")))
    self.add_option("Via Diameter", options.Dimension(self, defaultidx=2, 
                                               helptext=self.tr("Set via diameter")))
    self.add_option("Via Drill", options.Dimension(self, defaultidx=1, 
                                               helptext=self.tr("Set via drill size")))
    self.add_option("Via Mask", 
        options.Choice(self, ["Tented", "Drill", "Open"],
                       helptext=self.tr("Set via masking style")))
  
  def _findNearestEndpoint(self, pt, all=False):
    endpoints = self.allendpoints if all else self.endpoints
    mindist = 100000000000
    minidx = -1
    i = 0
    for pos, _, _ in endpoints:
      pdist = distPointToPoint(pos, pt)
      if pdist < mindist:
        mindist = pdist
        minidx = i
      i += 1
    return endpoints[minidx]

  def _focusLayer(self, layer):
    if config.get_config("general", "route_focus"):
      self.view().render.focusLayer(layer)

  def _isEndpoint(self, pt, all=False):
    eps = self.allendpoints if all else self.endpoints
    for pos, pad, foot in eps:
      if pos == pt:
        if pad.drill > 0:
          return True
        elif foot and (pad.getLayer(foot.mirror) == self.get_option("Layer")):
          return True
    return False

  def _mkvia(self, pt):
    view = self.view()
    for obj in view.design.findPoint(pt):
      if isinstance(obj, Via):
        return
    v = view.design.make_obj(Via, view.design, pt)
    v.width = v.height = self.get_option("Via Diameter")
    v.drill = self.get_option("Via Drill")
    vstyle = self.get_option("Via Mask")
    if vstyle == "Tented":
      v.stoptop = v.stopbot = v.drill
      v.stop = False
    elif vstyle == "Drill":
      v.stoptop = v.stopbot = v.drill
    elif vstyle == "Open":
      v.stoptop = v.stopbot = v.width
    v.name = self.netname

  def option_changed(self, optname, value):
    Tool.option_changed(self, optname, value)
    if optname == "Layer":
      view = self.view()
      pt = view.abs_pos_grid()
      if len(self.pts)>0:
        pt = self.pts[-1]
      if len(self.endpoints):
        if not self._isEndpoint(pt, True):
          self._mkvia(pt)
        self.move(0,0,0,0)

  def exit(self):
    while len(self.lines):
      t = self.lines.pop()
      t.delete()
    while len(self.ghosts):
      g = self.ghosts.pop()
      g.delete()
    self._focusLayer(None)

  def enter(self):
    self.set_help(self.tr("Click to route trace"))
    view = self.view()
    layout = view.design
    sel = view.selection.all()[0]
    self.netname = sel.name
    self.ghosts = []
    self._prevpt = None

    def getNamedConnector(part, connname):
      for pos, conn in part.connectors():
        if conn.name == connname:
          return conn, pos
      return None

    def getFootprint(name):
      fobjs = layout.findName(symname)
      for fobj in fobjs:
        if isinstance(fobj, SubPart):
          return fobj

    self.endpoints = [] # all pads
    self.allendpoints = [] # all pads and vias

    for symname, pinname in layout.project.conn.netpins(self.netname):
      foot = getFootprint(symname)
      if foot:
        pad, padpos = getNamedConnector(foot, pinname)
        self.endpoints.append((padpos, pad, foot))
        self.allendpoints.append((padpos, pad, foot))

    for v in layout.findName(self.netname):
      if isinstance(v, Via):
        self.allendpoints.append((v.pos, v, None))

    if len(self.endpoints) == 0:
      self.tools.show_message(
          "No net named '%s'" % self.netname,
          "No net named '%s' exists in the schematic. It may have been renamed\
           or deleted. Reroute or rename this trace first." % self.netname)
      self.tools.switchtool("Select")
      return
    
    mousepos = view.abs_pos()
    if isinstance(sel, Trace):
      self.set_option("Width", sel.width)
      self.set_option("Layer", sel.layer)
      if view.selection.point not in [0,1]:
        da = distPointToPoint(mousepos, sel.pta)
        db = distPointToPoint(mousepos, sel.ptb)
        pos = sel.pta if da < db else sel.ptb
      else:
        pos = sel.pta if view.selection.point == 0 else sel.ptb
    elif isinstance(sel, Via):
      pos = sel.pos
      for obj in layout.findPoint(pos):
        if isinstance(obj, Trace) and obj.name == self.netname:
          self.set_option("Width", obj.width)
          self.set_option("Layer", Layers.mirror(obj.layer))
          break
    elif isinstance(sel, Arc):
      self.set_option("Width", sel.width)
      self.set_option("Layer", sel.layer)
      pta,ptb = sel.points()
      da = distPointToPoint(mousepos, pta)
      db = distPointToPoint(mousepos, ptb)
      pos = pta if da < db else ptb
    else:
      pos, pad, foot = self._findNearestEndpoint(mousepos)
      self.set_option("Layer", pad.getLayer(foot.mirror))
      self.endpoints.remove((pos, pad, foot))
    view.render.setLayerVisible(self.get_option("Layer"), True)
    self._focusLayer(self.get_option("Layer"))
    self.pts = [pos]
  
  def _mkseg(self, pta, ptb, ghost=False, focus=True, name=None):
    view = self.view()
    s = view.design.make_obj(Trace, view.design, pta, ptb, ghost=ghost, name=name)
    s.layer = self.get_option("Layer")
    s.width = self.get_option("Width")
    if focus:
      self._focusLayer(self.get_option("Layer"))
    return s

  def _mkarc(self, pta, ptb, flip=False, ghost=False, focus=True, name=None):
    view = self.view()
    if flip:
      cx, cy = pta[0], ptb[1]
    else:
      cx, cy = ptb[0], pta[1]
    pos = (cx, cy)
    ofsa = pta[0]-cx, pta[1]-cy
    ofsb = ptb[0]-cx, ptb[1]-cy

    if (ofsb[0]-ofsa[0])*(ofsb[1]+ofsa[1]) > 0:
      ofsa, ofsb = ofsb, ofsa

    a = view.design.make_obj(Arc, view.design, pos, ofsa, ofsb, ghost=ghost, name=name)
    a.layer = self.get_option("Layer")
    a.width = self.get_option("Width")
    if focus:
      self._focusLayer(self.get_option("Layer"))
    return a

  def _splitseg(self, seg, pt):
    seg.delete()
    view = self.view()
    s1 = view.design.make_obj(Trace, view.design, seg.pta, pt, seg.name)
    s2 = view.design.make_obj(Trace, view.design, pt, seg.ptb, seg.name)
    s1.layer = s2.layer = seg.layer
    s1.width = s2.width = seg.width

  def _make_segs(self, pt):
    sx,sy, ex,ey = self.pts[0][0], self.pts[0][1], pt[0], pt[1]
    # trapazoidal points
    ldx, ldy = abs(sx-ex), abs(sy-ey)
    dix = 1 if (ex-sx) > 0 else -1
    diy = 1 if (ey-sy) > 0 else -1

    def trapa():
      if ldx > ldy:
        return ex-(ldy*dix), sy, 1
      else:
        return sx, ey-(ldx*diy), 0

    def trapb():
      if ldx > ldy:
        return sx+(ldy*dix), ey, 0
      else:
        return ex, sy+(ldx*diy), 1

    routetype = self.get_option("Style")
    if routetype == "Direct":
      return [("l", (sx,sy), (ex,ey), 0)]
    elif routetype == "Trapezoid A":
      mx,my,_ = trapa()
      if ldx == 0 or ldy == 0 or ldx == ldy:
        return [("l", (sx,sy), (ex,ey), 0)]
      return [("l", (sx,sy), (mx,my), 0), ("l", (mx,my), (ex,ey), 0)]
    elif routetype == "Trapezoid B":
      mx,my,_ = trapb()
      if ldx == 0 or ldy == 0 or ldx == ldy:
        return [("l", (sx,sy), (ex,ey), 0)]
      return [("l", (sx,sy), (mx,my), 0), ("l", (mx,my), (ex,ey), 0)]
    elif routetype == "Arc A":
      mx,my,flip = trapa()
      if ldx == 0 or ldy == 0:
        return [("l", (sx,sy), (ex,ey), 0)]
      elif ldx == ldy:
        return [("a", (sx,sy), (ex,ey), flip)]
      return [("l", (sx,sy), (mx,my), 0), ("a", (mx,my), (ex,ey), flip)]
    elif routetype == "Arc B":
      mx,my,flip = trapb()
      if ldx == 0 or ldy == 0:
        return [("l", (sx,sy), (ex,ey), 0)]
      elif ldx == ldy:
        return [("a", (sx,sy), (ex,ey), flip)]
      return [("a", (sx,sy), (mx,my), flip), ("l", (mx,my), (ex,ey), 0)]
    else:
      raise Exception()

  def move(self, x, y, dx, dy):
    view = self.view()
    pt = view.abs_pos_grid()
    a,b = pt
    while len(self.lines):
      t = self.lines.pop()
      t.delete()
    while len(self.ghosts):
      g = self.ghosts.pop()
      g.delete()

    # draw nearest line
    pos, _, _ = self._findNearestEndpoint(pt, all=True)
    t = view.render.addLines(Layers.BoundBox, [pt[0], pt[1], pos[0], pos[1]])
    self.lines.append(t)
    dia = self.get_option("Via Diameter")
    t = view.render.addLines(Layers.BoundBox, helpers.genLinePoints(
                             helpers.genCircle(pt, dia/2, 16)))
    self.lines.append(t)

    # draw route line
    segs = self._make_segs(pt)
    for st, pta, ptb, flip in segs:
      if st == "l":
        self.ghosts.append(self._mkseg(pta, ptb, ghost=True))
      elif st == "a":
        self.ghosts.append(self._mkarc(pta, ptb, flip, ghost=True))
    
    if config.get_config("general", "route_show_drc_bounds"):
      dsegs = []
      for _, pta, ptb, _ in segs:
        dsegs.extend(pta+ptb)
      self._drc(dsegs)

  def _drc(self, segs):
    dia = self.get_option("Via Diameter")
    spacing = self.drc.config("min_spacing")
    pt = self.view().abs_pos_grid()
    if pt != self._prevpt:
      self._prevpt = pt
      bb = BoundBox.fromPoints(segs)
      objs = self.view().design.findArea(bb.expand((spacing+dia)))
      self._checkobjs = []
      for o in objs:
        if (not o.ghost): 
          if isinstance(o, Trace) and o.name != self.netname:
            self._checkobjs.append(o)
          elif isinstance(o, Via):
            self._checkobjs.append(o)
          elif isinstance(o, Footprint):
            for e in o._getView():
              if isinstance(e, Pad):
                self._checkobjs.append(e)
    # draw outlines for objs nearby
    for o in self._checkobjs:
      if o.bb.ptdist(pt) < (spacing+dia):
        self.lines.append(self._drcerror(o, spacing))
  
  def _drcerror(self, obj, ofs=0):
    view = self.view()
    layer = Layers.HighlightDRC
    if isinstance(obj, Trace):
      lines = helpers.genSegment(obj.pta+obj.ptb, width=obj.width+ofs*2,
                                 capcount=8, outline=True)
    elif isinstance(obj, Pad):
      pos = obj._transform.points(obj.pos)
      if obj.shape == "circle":
        shape = helpers.genCircle(pos, obj.width/2+ofs, 16)
      elif obj.shape == "octagon":
        shape = helpers.genCircle(pos, obj.width/2+ofs, 8)
      elif obj.shape == "rectangle":
        w2, h2 = obj.width/2+ofs, obj.height/2+ofs
        pts = [-w2,-h2,w2,-h2,w2,h2,-w2,h2]
        shape = obj._transform.trans(*obj.pos).points(pts)
      else:
        shape = helpers.genCircle(pos, obj.width/2+ofs, 16)
      lines = helpers.genLinePoints(shape)
    
    return view.render.addLines(layer, lines)


  def _get_nearest_obj(self, pt, objs):
    mindist = 10000000
    nearest = npt = None
    for obj in objs:
      if isinstance(obj, Via) and obj.name == self.netname:
        if distPointToPoint(pt, obj.pos) < mindist:
          nearest, mindist, npt = obj, distPointToPoint(pt, obj.pos), obj.pos
      elif (isinstance(obj, Trace) and obj.name == self.netname and
            obj.layer == self.get_option("Layer")): 
        if distPointToLine(pt, obj.pta, obj.ptb) < mindist:
          lpt = nearestPointOnLine(pt, obj.pta, obj.ptb)
          nearest, mindist, npt = obj, distPointToPoint(pt, lpt), lpt
        if distPointToPoint(pt, obj.pta) < mindist:
          nearest, mindist, npt = obj, distPointToPoint(pt, obj.pta), obj.pta
        if distPointToPoint(pt, obj.ptb) < mindist:
          nearest, mindist, npt = obj, distPointToPoint(pt, obj.ptb), obj.ptb
    return nearest, npt

  def _align_route_point(self):
    view = self.view()
    mindist = min(units.mm(0.5), view.render.grid.scale*2)
    pt = view.abs_pos()
    # lock the point to nearby objects on this net
    npos, npad, nfoot = self._findNearestEndpoint(pt)
    padok = ((npad.drill > 0) or 
             (npad.getLayer(nfoot.mirror) == self.get_option("Layer")))
    if padok and (distPointToPoint(pt, npos) < mindist):
      # use an endpoint if we have one
      return npos
    else:
      # otherwise search nearby objects
      x1,x2 = pt[0]-mindist/2, pt[0]+mindist/2
      y1,y2 = pt[1]-mindist/2, pt[1]+mindist/2
      nearby = view.design.findArea(BoundBox(x1,y1,x2,y2))
      obj, npt = self._get_nearest_obj(pt, nearby)
      if obj is not None:
        if isinstance(obj, Trace) and (npt != obj.pta) and (npt != obj.ptb):
          self._splitseg(obj, npt)
        return npt
    # otherwise grid align mouse point
    return view.grid_align(*pt)
  
  def press(self, x, y, button, mod):
    if button == mouse.LEFT:
      # lock to nearest endpoint
      pt = self._align_route_point()
      if len(self.pts)==0 or self.pts[-1] != pt:
        self.pts.append(pt)
      if len(self.pts) == 2:
        pa, pb = self.pts
        segs = self._make_segs(pb)
        self.pts.pop(0)
        for st, pta, ptb, flip in segs:
          if st == "l":
            self._mkseg(pta, ptb, focus=False, name=self.netname)
          elif st == "a":
            self._mkarc(pta, ptb, flip, focus=False, name=self.netname)
      
      if self._isEndpoint(pt):
        self.tools.switchtool("Select")
    elif button == mouse.RIGHT:
      routetype = self.get_option("Style")
      if routetype == "Trapezoid A":
        self.set_option("Style", "Trapezoid B")
      elif routetype == "Trapezoid B":
        self.set_option("Style", "Trapezoid A")
      elif routetype == "Arc A":
        self.set_option("Style", "Arc B")
      elif routetype == "Arc B":
        self.set_option("Style", "Arc A")

  def valid_for(self, view, part):
    return (not part) and (view == "layout")
