# vim:set ts=2 sw=2 sts=2 et:
from rzn.inputdev import mouse, key
from .tool import Tool
from rzn.obj.objs import (SubPart, Airwire, Trace, Via, Connector, Arc)
from rzn.obj.layers import Layers
from rzn.obj.objs.boundbox import BoundBox
from rzn.obj import funcs, units
from rzn import config

class ToolSelect(Tool):
  name = "Select"

  def __init__(self, tools):
    Tool.__init__(self, tools)
    self.helptext = self.tr("Select objects")
    self.text = ""
    self._selrect = None
    self._double = False
    self._selectionRect = None
    self._selectionCycle = []

  def enter(self):
    self.set_help(self.tr("Click: select, Double-click: move, %s-click: next-nearest") %
                  self.view().mod_key_name(key.MOD_CTRL))
    self.selection()
    self._selectionRect = None
    self._selectionCycle = []
    self.pressPos = None

  def exit(self):
    self._deleteSelectionRect()

  def _deleteSelectionRect(self):
    if self._selrect:
      self._selrect.delete()
      self._selrect = None

  def _startSelectionRect(self, x, y):
    z = self.view().viewzoom()
    px, py = (x-self.view().ox) / z, (y-self.view().oy) / z
    self.pressPos = (px,py)
  
  def _updateSelectionRect(self, x, y):
    if not self.pressPos:
      return
    z = self.view().viewzoom()
    p0x, p0y = self.pressPos
    p1x, p1y = (x-self.view().ox) / z, (y-self.view().oy) / z
    self._selectionRect = BoundBox(min(p0x,p1x), min(p0y,p1y), 
                                  max(p0x,p1x), max(p0y,p1y))
    self._deleteSelectionRect()
    bb = self._selectionRect
    a,b,c,d = bb.x1, bb.y1, bb.x2, bb.y2
    self._selrect = self.view().render.addLines(Layers.Selection, [
                      a,b,c,b, c,b,c,d, c,d,a,d, a,d,a,b])
  
  def selection(self):
    self.tools.notify("selection_properties")
    self.tools.notify("selection_actions")

  def valid_actions(self):
    r = Tool.valid_actions(self)
    sel = self.view().selection.all()
    validtools = []
    if len(sel) == 1:
      if isinstance(sel[0], (Airwire, Trace, Via)):
        validtools.append("Route")
      elif isinstance(sel[0], Arc):
        if len(self.view().design.project.conn.netpins(sel[0].name)):
          validtools.append("Route")
      elif isinstance(sel[0], (SubPart, Connector)):
        validtools.append("Duplicate")
    if len(sel) >= 1:
      validtools.append("Move")
    return r + validtools

  def _smart_nearest(self):
    # sort objects by prioritised type/layer
    def psort(objs):
      priority = [SubPart, Via, Airwire, Trace]
      def ti(obj):
        for i, t in enumerate(priority):
          if isinstance(obj, t):
            return i
        return len(priority)
      # comapre by type priority then layer
      def ocmp(a, b):
        r = cmp(ti(a), ti(b))
        if r == 0:
          if a.hasprop("layer") and b.hasprop("layer"):
            return Layers.zorder_cmp(a.layer, b.layer)
        return r
      return sorted(objs, ocmp)

    view = self.view()
    # maximum distance in pixels to select objects
    pix = 0.5 if view.name == "schematic" else 4
    max_dist = view.pixel_units(pix)
    pos = view.abs_pos()
    # search for a few objects and filter & sort to find the best selection
    layers = view.render.layers
    def layerfilter(o):
      return o.layersVisible(layers)
    item, items = view.design.findNearest(*pos, n=10, filterfn=layerfilter)
    if len(items):
      choices = [items[0]]
      d = items[0].distToPoint(pos) + max_dist
      for item in items[1:]:
        if item.distToPoint(pos) < d:
          choices.append(item)
      choices = psort(choices)
      if len(choices):
        return choices[0]
    return item

  def _nearest(self, mod):
    view = self.view()
    pos = view.abs_pos()
    if config.get_config("general", "priority_select", True):
      item = self._smart_nearest()
    else:
      item, _ = view.design.findNearest(*pos)
    # handle selection cycle
    if (mod & key.MOD_CTRL):
      if (mod & key.MOD_SHIFT):
        if len(self._selectionCycle) > 1:
          self._selectionCycle.pop()
          item = self._selectionCycle[-1]
      else:
        n = len(self._selectionCycle) + 1
        _, nearests = view.design.findNearest(*pos, n=n)
        for nearest in nearests:
          if nearest not in self._selectionCycle:
            self._selectionCycle.append(nearest)
            item = nearest
            break
    else:
      self._selectionCycle = [item]
    point = funcs.get_nearest_endpoint(item, pos)
    return [item],point

  def press(self, x, y, button, mod):
    self._double = False
    if button == mouse.LEFT:
      self._startSelectionRect(x, y)

  def doubleclick(self, x, y, button, mod):
    self._double = True

  def release(self, x, y, button, mod):
    view = self.view()
    olditems = view.selection.all()
    if button == mouse.LEFT:
      if not self.pressPos:
        return
      items = point = None
      if self._selectionRect:
        # if it's a tiny selection just do nearest search
        if (self._selectionRect.area()*(view.viewzoom()**2)) > 10:
          items = view.design.findArea(self._selectionRect)
      # otherwise nearest search
      if items is None:
        items, point = self._nearest(mod)
        if (items and items[0] and
            (items[0].distToPoint(view.abs_pos()) > units.mm(2)) and 
            (not items[0].bb.pointIn(view.abs_pos()))):
          items = []

      # update the selection
      if (mod & key.MOD_SHIFT) and not (mod & key.MOD_CTRL):
        view.selection.add(items)
      else:
        view.selection.set(items, point)
        # double click to move current selection
        if (olditems == items) and len(items) and (self._double):
          if not view.is_readonly():
            self.tools.switchtool("Move")

      self.selection()
      self._selectionRect = None
      self._deleteSelectionRect()
    elif button == mouse.RIGHT:
      view = self.view()
      if not view.is_readonly():
        popups = view.selection.actions.for_popup()
        self.tools.show_context_menu(x, y, popups)

  def drag(self, x, y, dx, dy, button, mod):
    if button & mouse.LEFT:
      self._updateSelectionRect(x, y)

  def cancel(self):
    self.view().selectObjects([])
    self.text = ""
    Tool.cancel(self)
