# vim:set ts=2 sw=2 sts=2 et:
from rzn.api import obj, Transform, Plugin, Logger, APIException, Layers, units
from collections import defaultdict
import sys, os, math

log = Logger(__name__)


class SvgLayer(object):
  def __init__(self):
    self.objs = []


class SvgExport(object):
  def __init__(self, viewname, config, options, layercount=0):
    self.viewname = viewname
    self.config = config
    self.layercount = layercount
    self.layers = defaultdict(lambda: SvgLayer())
    self.minx = self.miny = sys.maxint
    self.maxx = self.maxy = -sys.maxint
    self.colours = {}
    if not bool(options.get("uicolours", 0)):
      for layername, col in self.config.get("colours", {}).items():
        if hasattr(Layers, layername):
          self.colours[getattr(Layers, layername)] = col
    self.netdots = defaultdict(list)

  def setBounds(self, b):
    self.minx, self.miny, self.maxx, self.maxy = b

  def _colour(self, layer):
    if layer in self.colours:
      return self.colours[layer]
    else:
      return Layers.colours[layer]

  def _getbounds(self, pts):
    for i in range(len(pts)//2):
      x,y = pts[i*2+0], pts[i*2+1]
      if x < self.minx: self.minx = x
      if x > self.maxx: self.maxx = x
      if y < self.miny: self.miny = y
      if y > self.maxy: self.maxy = y

  def _layer(self, razenlayer):
    return self.layers[razenlayer]

  def _render(self, layout, transform, side=0, ignores=[]):
    def add_obj(layer, o):
      self._layer(layer).objs.append(o)

    def add_stop_obj(layer, o):
      if layer == Layers.CopperTop:
        add_obj(Layers.StopTop, o)
      elif layer == Layers.CopperBottom:
        add_obj(Layers.StopBottom, o)

    for e in layout:
      if e in ignores:
        continue
      if isinstance(e, obj.Segment):
        pts = transform.points(e.pta+e.ptb)
        self._getbounds(pts)
        add_obj(e.getLayer(side), ("line", e.width, pts))
        if isinstance(e, obj.Net):
          self.netdots[tuple(pts[:2])].append(e)
          self.netdots[tuple(pts[2:])].append(e)
      elif isinstance(e, obj.Arc):
        pts = e.genSegmented()
        self._getbounds(pts)
        for i in range((len(pts)/2)-1):
          seg = pts[i*2:i*2+4]
          add_obj(e.getLayer(side), ("line", e.width, seg))
      elif isinstance(e, obj.SubPart):
        t = transform.trans(*e.pos)
        fside = 1 if e.mirror else 0
        if e.mirror:
          t = t.flipx()
        t = t.rot(e.angle)
        v = e.epart.layout if self.viewname == "layout" else e.epart.schematic
        self._render(v, t, fside, [] if e.showname else [e.nameText()])
      elif isinstance(e, obj.Pad):
        pts = transform.points(e.pos)
        self._getbounds(pts)
        # pad
        if e.shape == "circle":
          oap, ow, oh, opts = "circle", e.width, e.height, pts
        elif e.shape == "octagon":
          ta = int(transform.angle())
          w2,h2 = e.width/2, e.height/2
          bpts = []
          da = 2*math.pi / 8.0
          ofs = 2*math.pi / 16.0
          for i in range(8):
            bpts.append(math.cos(i*da+ofs)*w2)
            bpts.append(math.sin(i*da+ofs)*h2)
          txf = Transform.ident().trans(*pts).rot(ta)
          tpts = txf.points(bpts)
          oap, ow, oh, opts = "polygon", e.width, e.height, tpts
        elif e.shape == "rect" or e.shape == "rectangle":
          ta = int(transform.angle())
          if (ta % 180) == 0:
            oap, ow, oh, opts = "rectangle", e.width, e.height, pts
          elif (ta % 180) == 90:
            oap, ow, oh, opts = "rectangle", e.height, e.width, pts
          else:
            rw2,rh2 = e.width/2, e.height/2
            rpts = [-rw2,-rh2, +rw2,-rh2, +rw2,+rh2, -rw2,+rh2]
            txf = Transform.ident().trans(*pts).rot(ta)
            tpts = txf.points(rpts)
            oap, ow, oh, opts = "polygon", e.width, e.height, tpts
        elif e.shape == "polygon":
          ta = int(transform.angle())
          txf = Transform.ident().trans(*pts).rot(ta)
          tpts = txf.points(e.polygon)
          oap, ow, oh, opts = "polygon", e.width, e.height, tpts
        else:
          raise APIException("unhandled pad shape: %s" % e.shape)
        # add points for copper and stop layers
        layer = e.getLayer(side)
        add_obj(layer, (oap, ow, oh, opts))
        add_stop_obj(layer, (oap, ow, oh, opts))
        edrill = e.properties.get("drill", e.width/2)
        if edrill > 0:
          for olayer in Layers.get("layout", "copper", coppercount=self.layercount):
            if olayer != layer:
              add_obj(olayer, (oap, ow, oh, opts))
          add_obj(Layers.Drill, ("circle", edrill, edrill, pts))
          add_stop_obj(Layers.mirror(layer), (oap, ow, oh, opts))
      elif isinstance(e, obj.Polygon):
        layer = e.getLayer(side)

        space = max(e.space, units.mm(0.05))
        # don't clip if we have no net set
        if e.net == "":
          clipped = [e.getCurrentTransform().points(e.pts)]
        else:
          clipped = e.clip(space+e.width/2, e.width/2, cliptype=obj.Polygon.CLIP_ACCURATE)
        add_obj(layer, ("hpolygon", 0, 0, clipped))
        # handle thermals
        if e.thermal > 0:
          for t in e.thermalpoints:
            add_obj(layer, ("line", e.thermal, t))
      elif isinstance(e, obj.Drill):
        edrill = e.properties["diameter"]
        pts = transform.points(e.pos)
        self._getbounds(pts)
        add_obj(Layers.Drill, ("circle", edrill, edrill, pts))
      elif isinstance(e, obj.Pin):
        pl = e.length
        w = obj.Pin.WIDTH
        txf = transform.trans(*e.pos).rot(e.angle)
        px, py = txf.points((0,0))
        pxp, pyp = txf.points((-pl,0))
        x1,x2 = px-w, px+w
        y1,y2 = py-w, py+w
        pts = [px,py,pxp,pyp]
        self._getbounds(pts)
        layer = e.getLayer(side)
        for i in range(len(pts)/4):
          add_obj(Layers.Symbol, ("line", 0, pts[i*4:i*4+4]))
        if e.showname:
          fnt = obj.Font.get("vector")
          txt = e._evalname()
          fw,fh = fnt.textSize(txt, 1)
          txtpos = txf.trans(-pl, 0).rot(-e.angle)
          tw = (fw + units.mm(0.5)) if e.angle < 180 else (-units.mm(0.5))
          rangle = e.angle % 180
          txtpos = txtpos.rot(rangle).trans(-tw, fh/2.0)
          self._rendertext(txt, fnt, 1, txtpos, layer, 200)
      elif isinstance(e, obj.Text):
        size = e.properties["size"]
        layer = e.getLayer(side)
        txt = e.evaluatedValue()
        self._rendertext(txt, e.getFont(), size, e.getCurrentTransform(), 
                         layer, e.charspace)
      else:
        raise APIException("unknown object: %s" % e)

  def _rendertext(self, text, font, size, transform, layer, charspace):
    def add_obj(layer, o):
      self._layer(layer).objs.append(o)

    w,h = font.textSize(text, size, charspacing=charspace)
    lines = font.genText(text, size, transform, charspacing=charspace)
    self._getbounds(lines)
    for i in range(len(lines)/4):
      add_obj(layer, ("line", units.mm(0.1), lines[i*4:(i*4)+4]))

  def _emits(self, s):
    self.of.write("%s\n" % s)

  def _emito(self, n, end=True, **a):
    args = " ".join("%s=\"%s\"" % (x.replace("_","-"),y) for x,y in a.items())
    self.of.write("<%s %s%s>\n" % (n, args, '/' if end else ''))

  def _genLayer(self, lname, layer):
    def cx(v): return v-self.minx+self.pad
    def cy(v): return v-self.miny+self.pad
    def cp(pts):
      r = []
      for i in range(len(pts)/2):
        r.append(pts[i*2+0]-self.minx+self.pad)
        r.append(pts[i*2+1]-self.miny+self.pad)
      return r

    col = (",".join([str(int(x*255)) for x in self._colour(lname)[:3]]))
    alpha = "%0.2f" % self._colour(lname)[3]
    self._emito("g", end=False, stroke="rgb(%s)" % col,stroke_opacity=alpha,
                                fill="rgb(%s)" % col, fill_opacity=alpha,
                                stroke_linecap="round")
    for obj in layer.objs:
      ot = obj[0]
      oa = obj[1:]
      if ot == "line":
        w, pts = oa
        ve = ""
        if w == 0:
          if lname == Layers.Border:
            w = units.mm(0.1)
          else:
            w = units.mm(0.1)
            #w = "" #50
            #ve = "non-scaling-stroke"
        self._emito("line", x1=cx(pts[0]), y1=cy(pts[1]),
                            x2=cx(pts[2]), y2=cy(pts[3]),
                            stroke_width=w, vector_effect=ve)
      elif ot == "circle":
        w, h, pos = oa
        self._emito("ellipse", rx=w/2, ry=h/2, cx=cx(pos[0]), cy=cy(pos[1]))
      elif ot == "rectangle":
        w, h, pos = oa
        self._emito("rect", width=w, height=h, 
                            x=cx(pos[0]-w/2), y=cy(pos[1]-h/2))
      elif ot == "polygon":
        w, h, pts = oa
        self._emito("polygon", points=",".join(map(str, cp(pts))))
      elif ot == "hpolygon":
        _, _, polys = oa
        pstr = ""
        for poly in polys:
          pts = cp(poly)
          for i in range(len(pts)/2):
            pcmd = "M" if i == 0 else "L"
            pstr += "%s%d,%d " % (pcmd, pts[i*2], pts[i*2+1])
        self._emito("path", d=pstr, fill_rule="nonzero")
      else:
        print "unhandled", ot, oa
    self._emits("</g>")

  def save(self, objs, fn, splitlayers=False):
    self._render(objs, Transform.ident())
    # render net dots
    for pt, nets in self.netdots.items():
      if len(nets) > 2:
        sz = units.mm(0.5)
        self._layer(Layers.Net).objs.append(("circle", sz, sz, pt))

    self.pad = units.mm(1)
    def emit_head():
      self._emits("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>")
      vw = self.maxx-self.minx + self.pad*2
      vh = self.maxy-self.miny + self.pad*2
      vb = "%d %d %d %d" % (0,0,vw,vh)
      w = "%dmm" % (vw/units.mm(1.))
      h = "%dmm" % (vh/units.mm(1.))
      self._emits("<svg xmlns=\"http://www.w3.org/2000/svg\" version=\"1.1\" viewBox=\"%s\" width=\"%s\" height=\"%s\">" % (vb, w, h))
      self._emits("<rect width=\"%s\" height=\"%s\" style=\"fill:rgb(255,255,255);\"/>" % (vw, vh))
    def emit_foot():
      self._emits("</svg>")

    if splitlayers:
      for lname in Layers.stackup:
        if lname in self.layers:
          self.of = open("%s_%02d.svg" % (fn, lname), "wb")
          emit_head()
          self._genLayer(lname, self.layers[lname])
          emit_foot()
    else:
      self.of = open("%s.svg" % fn, "wb")
      emit_head()
      for lname in Layers.stackup:
        if lname in self.layers:
          self._genLayer(lname, self.layers[lname])
      emit_foot()


class SvgExportPlugin(Plugin):
  def __init__(self):
    Plugin.__init__(self, "SVG", Plugin.EXPORT)

  def main(self, ui):
    self.batchMain(ui.currentTab().project(), {})

  def batchMain(self, proj, options):
    basefn = options.get("fn")
    if basefn is None:
      basefn = os.path.join(proj.filename(), 
                            os.path.basename(proj.filename()))
    view = options.get("view")
    lsplit = bool(options.get("layersplit",0))
    if view == "layout" or view is None:
      layercount = proj.copperlayers()
      e = SvgExport("layout", self.loadConfig("svg"), options, layercount)
      e.setBounds(proj.layout.viewBounds())
      e.save(proj.layout.objs(), basefn+"_brd", lsplit)
    if view == "schematic" or view is None:
      e = SvgExport("schematic", self.loadConfig("svg"), options)
      e.setBounds(proj.schematic.viewBounds())
      e.save(proj.schematic.objs(), basefn+"_sch", lsplit)
