# vim:set ts=2 sw=2 sts=2 et:
from rzn.api import GeneratorPlugin, Layers
from rzn.api.pluginoption import IntOption, FloatOption
from rzn.api.units import mm
from rzn.api.common import addrect, markerdot

class SOICPlugin(GeneratorPlugin):
  def __init__(self):
    GeneratorPlugin.__init__(self, "SOIC/SOP", [
      ("Pin count",       "count",  IntOption(8)),
      ("Pin Pitch (mm)",  "pitch",  FloatOption(1.27)),
      ("Pin length (mm)", "pinw",   FloatOption(1.5)),
      ("Pin height (mm)", "pinh",   FloatOption(0.6)),
      ("Body width (mm)", "width",  FloatOption(3.81)),
    ],[
      ("SOIC8",    [ 8, 1.27, 1.5, 0.6, 3.81]),
      ("SOIC8-W",  [ 8, 1.27, 1.5, 0.6, 7.62]),
      ("SOIC14",   [14, 1.27, 1.5, 0.6, 3.81]),
      ("SOIC14-W", [14, 1.27, 1.5, 0.6, 7.62]),
      ("SOIC16",   [16, 1.27, 1.5, 0.6, 3.81]),
      ("SOIC16-W", [16, 1.27, 1.5, 0.6, 7.62]),
      ("SSOP14",   [14, 0.65, 1.5, 0.4, 5.30]),
      ("SSOP16",   [16, 0.65, 1.5, 0.4, 5.30]),
      ("SSOP20",   [16, 0.65, 1.5, 0.4, 5.30]),
    ])

  def footprint(self, project, options):
    pins, width = options["count"], options["width"]
    pinlength, pinheight = options["pinw"], options["pinh"]
    pitch = options["pitch"] * mm(1)
    width *= mm(1)
    pinlength *= mm(1)
    pinheight *= mm(1)
    pps = pins / 2 # pins per side
    bodyheight = (pps-1)*pitch + mm(1.27)
    po = ((pps-1)*pitch)/2.0
    project.grid(1)

    # add a rectangular pad
    def addpad(n, x, y):
      p=project.layout.pad((x,y), pinlength, pinheight, "rectangle")
      p.name = "%d" % (n+1)

    # add pads
    for i in range(pps):
      addpad(i,       -(width+pinlength)/2.0, i*pitch-po)
      addpad(pins-i-1, (width+pinlength)/2.0, i*pitch-po)

    # add body outline
    addrect(project.layout.line, width/2.0, bodyheight/2.0, mm(0.25), 
            Layers.OutlineTop)
    markerdot(project.layout, -width/2.0-mm(0.6), -bodyheight/2.0-mm(0.6), 
              mm(0.3))
    # name
    project.layout.text((-width/2.0,-bodyheight/2.0-mm(0.1)), 
                        "$name", 1, Layers.NameTop)

  def symbol(self, project, options):
    pins = options["count"]
    pps = pins / 2 # pins per side
    height = (pps+1)*2
    width = 16
    po = (height/2.0 - 2)
    project.grid(mm(1.27))
    
    # add a pin
    def addpin(n, x, y):
      p = project.schematic.pin((x,y))
      p.name = "%d" % (n+1)
      p.angle = 180 if x < 0 else 0

    # add pins
    for i in range(pps):
      addpin(i,       -width/2-2, i*2-po)
      addpin(pins-i-1, width/2+2, i*2-po)

    # add body outline
    addrect(project.schematic.line, width/2.0, height/2.0, 0, Layers.Symbol)
    # name
    project.schematic.text((-width/2.0,-height/2.0), "$name", 1, Layers.Label)
