# vim:set ts=2 sw=2 sts=2 et:
from rzn.api import Logger, VCPluginFile, VCNotFound
import os
import json
from . import raw, hgrc

log = Logger(__name__)


class HgPluginFile(VCPluginFile):
  def __init__(self, vcplugin, basedir, fn):
    VCPluginFile.__init__(self, vcplugin, basedir, fn)
    self._revcache = {}
    if not os.path.exists(os.path.join(basedir, ".hg")):
      assert "Repo doesn't exist"

  def _statuscheck(self, commit=True):
    stat,_ = self._hgr("status", self.path())
    if stat.startswith("?"):
      log.info("%s: not in repo, adding" % self)
      self._hgr("add", self.path())
      if commit:
        self.commit("initial import of %s" % self.fn)

  def uncommited(self):
    stat,_ = self._hgr("status", self.path())
    for l in stat.split("\n"):
      if l.startswith("M"):
        return True
    return False

  def discardUncommitted(self):
    self._hgr("up", "-C")

  def checkUsername(self):
    if hgrc.read(self.basedir, "ui", "username") is None:
      log.info("No username set")
      self.vcplugin.setUsername(self.basedir)

  def _cat(self, rev):
    while 1:
      js, stat = self._hgrn("cat", "-r", str(rev), self.path())
      if js == "":
        log.error("Bad JSON for %s:%s (%d)" % (self.path(), rev, stat))
      else:
        return js, stat

  def load(self, rev):
    if rev == None: # load the latest
      return json.load(file(self.path()))
    elif rev in self._revcache:
      try:
        return json.loads(self._revcache[rev])
      except ValueError:
        raise Exception("can't load cached JSON: '%s' for file %s" % (
                        self._revcache[rev], self.path())) 
    else: # load specific revision
      js, stat = self._cat(rev)
      if stat != 0:
        if "no such file" in js:
          raise  VCNotFound("%s (rev %s) not found" % (self.path(), str(rev)))
        else:
          raise Exception("Can't load %s rev %s: %d" % (self.fn, rev, stat))
      if js == "":
        raise Exception("empty JSON revcache store %s:%s" % (self.path(), rev))
      self._revcache[rev] = js
      return json.loads(js)

  def save(self, js, commit, msg, statusCheck=True):
    f = file(self.path(), "wb")
    json.dump(js, f)
    f.close()
    if statusCheck:
      self._statuscheck(commit)
    if commit:
      self.commit(msg)

  def saveraw(self, data, commit, msg):
    f = file(self.path(), "wb")
    f.write(data.encode("utf-8"))
    f.close()
    self._statuscheck(commit)
    if commit:
      self.commit(msg)

  def remove(self, commit=False):
    self._hgr("remove", self.path())
    if commit:
      self.commit("removed %s" % self.fn)

  def commit(self, msg):
    self.checkUsername()
    if not msg:
      msg = "save %s" % self.fn
    args = ["commit", "-m", msg]
    self._hgr(*args)

  def history(self):
    hist, _ = self._hgr("log", self.path())
    return raw.parse_hglog(hist)

  def status(self, rev=None):
    hist = ""
    # FIXME: hack around issue #39 - strange mac hg issue
    while hist == "":
      if rev:
        hist, _ = self._hgr("log", "--debug", "-r", str(rev), self.path())
      else:
        hist, _ = self._hgr("log", "--debug", "-l", "1", self.path())
      if hist == "":
        log.error("Bad history for '%s'" % self.path())
    return raw.parse_hglog(hist)[0]

  def rev(self):
    return self.status().get("changeset")

  def haveRevision(self, rev):
    try:
      self._hgr("log", "-r", str(rev), self.path())
      return True
    except:
      return False

  def exists(self, rev):
    if rev == None:
      return os.path.exists(self.path())
    js, stat = self._cat(rev)
    if (stat == 0):
      self._revcache[rev] = js
    return stat == 0

  def parentrevs(self, rev):
    changes, _ = self._hgr("parents", "-r", str(rev))
    log = raw.parse_hglog(changes)
    if len(log) == 1 and "changeset" in log[0]:
      return log[0]["rev"]
    return -1
  
  def push(self, url=None):
    if self.remoteUrl():
      args = ["push"]
      if url:
        args.append(url)
      self._hgr(*args)
  
  def pull(self):
    if self.remoteUrl():
      self._hgr("pull", "-u")

  def outgoing(self, url=None):
    if self.remoteUrl():
      args = ["outgoing"]
      if url:
        args.append(url)
      changes, _ = self._hgr(*args)
      return list(reversed(raw.parse_hglog(changes)))
    return []
  
  def incoming(self):
    if self.remoteUrl():
      changes, _ = self._hgr("incoming")
      return list(reversed(raw.parse_hglog(changes)))
    return []

  def remoteUrl(self):
    return hgrc.read(self.basedir, "paths", "default")

  def setRemoteUrl(self, url):
    hgrc.write(self.basedir, "paths", "default", url)

  def _hgr(self, *opts):
    out,rc = raw.hg("-R", self.basedir, *opts)
    if "abort: " in out:
      raise Exception("HG Error: %s" % out)
    return out,rc

  def _hgrn(self, *opts):
    return raw.hg("-R", self.basedir, *opts)

